#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <poll.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <SDL/SDL.h>
#include "macros.h"
#include "a_oscill.h"

/* structure for find color from string*/
struct a_oscill_color {
	char *namecolor;
	int numcolor;
} color_list[] = {
		{"red", red},
		{"blue", blue},
		{"yellow", yellow},
		{"green", green}
       };


/*--------------------function for circle_buff: begin--------------*/
struct circle_buff *
a_oscill_new_circle_buff(int buff_size)
{

	struct circle_buff *c_buff;

	c_buff = malloc(sizeof(struct circle_buff));

	if (c_buff == NULL)
		return NULL;

	c_buff->buff = malloc(sizeof(short) * buff_size);

	if (c_buff->buff == NULL) {
		free(c_buff);
		return NULL;
	}	

	c_buff->head      = 0;
	c_buff->tail      = 0;
	c_buff->buff_size = buff_size;

	return c_buff;
}

void
a_oscill_write_c_buff(struct circle_buff *c_buff, short sample)
{

	return_if_fail(c_buff != NULL);

	return_if_fail(c_buff->buff != NULL);

	c_buff->buff[c_buff->head] = sample;

	c_buff->head = (c_buff->head + 1) % c_buff->buff_size;

	if (c_buff->head == c_buff->tail)
		c_buff->tail = (c_buff->tail + 1) % c_buff->buff_size;

}

short
a_oscill_read_c_buff(struct circle_buff *c_buff)
{
	short sample;

	return_val_if_fail(c_buff != NULL, A_OSCILL_ERR);

	return_val_if_fail(c_buff->buff != NULL, A_OSCILL_ERR);

	return_val_if_fail(c_buff->tail != c_buff->head, A_OSCILL_ERR);

	sample = c_buff->buff[c_buff->tail];

	c_buff->tail = (c_buff->tail + 1) % c_buff->buff_size;

	return sample;

}

void
a_oscill_free_c_buff(struct circle_buff **c_buff)
{
	free((*c_buff)->buff);
	free(*c_buff);

	*c_buff = NULL;
}

/*--------------------function for circle_buff: end----------------*/


/*------------------------function for screen: begin-----------------*/
void
a_oscill_slock(SDL_Surface *screen)
{
	if (SDL_MUSTLOCK(screen))
		if (SDL_LockSurface(screen) < 0)
			return;
}

/* ------------------------------------------------------ */
void
a_oscill_sulock(SDL_Surface *screen)
{
	if (SDL_MUSTLOCK(screen))
		SDL_UnlockSurface(screen);
}

/* ------------------------------------------------------ */
void
a_oscill_DrawPixel(SDL_Surface *screen, int x, int y, Uint8 R, Uint8 G, Uint8 B)
{
	return_if_fail(x < screen->w);
	return_if_fail(y < screen->h);

	Uint32 color = SDL_MapRGB(screen->format, R, G, B);
	switch (screen->format->BytesPerPixel) {
	case 1: { // Assuming 8-bpp
		Uint8 *bufp;
		bufp = (Uint8 *)screen->pixels + y * screen->pitch + x;
		*bufp = color;
	} break;
	case 2: {// Probably 15-bpp or 16-bpp
		Uint16 *bufp;
		bufp = (Uint16 *)screen->pixels + y * screen->pitch / 2 + x;
		*bufp = color;
	} break;
	case 3: {// Slow 24-bpp mode, usually not used
		Uint8 *bufp;
		bufp = (Uint8 *)screen->pixels + y * screen->pitch + x * 3;
		if(SDL_BYTEORDER == SDL_LIL_ENDIAN) {
			bufp[0] = color;
			bufp[1] = color >> 8;
			bufp[2] = color >> 16;
		} else {
			bufp[2] = color;
			bufp[1] = color >> 8;
			bufp[0] = color >> 16;
		}
	} break;
	case 4: {// Probably 32-bpp
		Uint32 *bufp;
		bufp = (Uint32 *)screen->pixels + y * screen->pitch / 4 + x;
		*bufp = color;
	} break;
	}

}

void
a_oscill_draw_screen(SDL_Surface *screen, int screencolor)
{

	int i, j;
	int r, g, b;

	switch (screencolor) {
	case red:
		r = 240;
		g = 0;
		b = 0;
		break;

	case blue:
		r = 110;
		g = 202;
		b = 239;
		break;

	case green:
		r = 0;
		g = 240;
		b = 0;
		break;

	case yellow:
		r = 246;
		g = 248;
		b = 94;
		break;
	default:
		r = 255;
		g = 255;
		b = 255;
		break;
	}

	a_oscill_slock(screen);

	for (i = 0; i < SCREEN_WIDTH; i++)
		for (j = 0; j < SCREEN_HEIGHT; j++)
			if (i % CELL_PIXELS_WIDTH != 0 && j % CELL_PIXELS_HEIGHT != 0)
				a_oscill_DrawPixel(screen, i, j, r, g, b);
			else
				a_oscill_DrawPixel(screen, i, j, 0, 0, 0);

	for (i = -NUM_PIXEL_DIV / 2; i <= NUM_PIXEL_DIV / 2; i++)
		for (j = 0; j < SCREEN_HEIGHT; j++)
			if (j % NUM_PIXEL_DIV == 0)
				a_oscill_DrawPixel(screen, SCREEN_WIDTH / 2 + i, j, 0, 0, 0);

	for (j = -NUM_PIXEL_DIV / 2; j <= NUM_PIXEL_DIV / 2; j++)
		for (i = 0; i < SCREEN_WIDTH; i++)
			if (i % NUM_PIXEL_DIV == 0)
				a_oscill_DrawPixel(screen, i, SCREEN_HEIGHT / 2 + j, 0, 0, 0);

	a_oscill_sulock(screen);

}

void
a_oscill_update_screen(SDL_Surface *screen, struct circle_buff *c_b, int ccolor, int scolor)
{
	return_if_fail(c_b != NULL);
	return_if_fail(c_b->buff != NULL);

	int count, i, r, g, b;
	short data, prevdata;

	switch (ccolor) {
		case red:
			r = 240;
			g = 0;
			b = 0;
			break;

		case blue:
			r = 110;
			g = 202;
			b = 239;
			break;

		case green:
			r = 0;
			g = 240;
			b = 0;
			break;

		case yellow:
			r = 246;
			g = 248;
			b = 94;
			break;
		default:
			r = 0;
			g = 0;
			b = 0;
			break;
		}

	a_oscill_draw_screen(screen, scolor);

	a_oscill_slock(screen);

	count = 0;

	if (c_b->head != c_b->tail) {
		prevdata = a_oscill_read_c_buff(c_b);
		if (prevdata > -1 && prevdata < SCREEN_HEIGHT)
			a_oscill_DrawPixel(screen, count, prevdata, r, g, b);
		count++;
	} else
		return;

	while (count < SCREEN_WIDTH && c_b->tail != c_b->head ) {

		data = a_oscill_read_c_buff(c_b);

		if (data < prevdata)
			for (i = data; i <= prevdata; i++)
				if (i > -1 && i < SCREEN_HEIGHT)
					a_oscill_DrawPixel(screen, count-1, i, r, g, b);

		if (data > prevdata)
			for (i = prevdata; i <= data; i++)
				if (i >-1 && i < SCREEN_HEIGHT)
					a_oscill_DrawPixel(screen, count-1, i, r, g, b);

		if (data > -1 && data < SCREEN_HEIGHT)
			a_oscill_DrawPixel(screen, count, data, r, g, b);

		prevdata = data;
		count++;
	}

	a_oscill_sulock(screen);
	SDL_Flip(screen);
}

/*------------------------function for screen: end-------------------*/

/*------------------------function for oscill: begin-----------------*/

/* Destroy struct a_oscill. */
void
a_oscill_free(struct a_oscill **a_oscill)
{
	if ((*a_oscill)->screen != NULL)
		SDL_FreeSurface((*a_oscill)->screen);

	a_oscill_free_c_buff(&(*a_oscill)->c_buff);


	free(*a_oscill);

	(*a_oscill) = NULL;
}

struct a_oscill *
a_oscill_new(int screencolor, int curvecolor)
{
	struct a_oscill *oscill;
	int res;

	oscill = malloc(sizeof(*oscill));

	if (oscill == NULL)
		return NULL;

	memset(oscill, 0, sizeof(*oscill));

	oscill->scolor     = screencolor;
	oscill->ccolor     = curvecolor;
	oscill->volt_div   = 0.001;
	oscill->time_div   = 0.0000001;
	oscill->i_time_div = 0.0000001;
	oscill->solve_time = 0.0;
	oscill->ndata      = 0;
	oscill->flag       = 0;

	res = SDL_Init(SDL_INIT_VIDEO);

	if (res < 0) {
		fprintf(stderr, "unable to init SDL: %s\n", SDL_GetError());
		goto err_init;
	}

	oscill->screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 32, SDL_HWSURFACE | SDL_RESIZABLE);

	if (oscill->screen == NULL) {
		fprintf(stderr, "can't set videomode: %s\n",SDL_GetError());
		goto err_video;
	}

	oscill->c_buff = a_oscill_new_circle_buff(2 * SCREEN_WIDTH);

	if (oscill->c_buff == NULL) {
		fprintf(stderr, "can't allocate cbuf\n");
		goto err_cbuf;
	}

	a_oscill_draw_screen(oscill->screen, screencolor);

	oscill->a_oscill_update_screen = a_oscill_update_screen;
	oscill->a_oscill_deinit        = a_oscill_free;
	atexit(SDL_Quit);

	return oscill;
err_cbuf:

err_video:
	SDL_Quit();
err_init:
	free(oscill);
	return NULL;
}

void
a_oscill_set_time_div(struct a_oscill *oscill, double time_div)
{
	return_if_fail(oscill != NULL);

	oscill->time_div = time_div;

	oscill->c_buff->head = 0;
	oscill->c_buff->tail = 0;
}

void
a_oscill_set_i_time_div(struct a_oscill *oscill, double i_time_div)
{
	return_if_fail(oscill != NULL);

	oscill->i_time_div = i_time_div;

	oscill->c_buff->head = 0;
	oscill->c_buff->tail = 0;

}

void
a_oscill_set_volt_div(struct a_oscill *oscill, double volt_div)
{
	return_if_fail(oscill != NULL);

	oscill->volt_div = volt_div;

	oscill->c_buff->head = 0;
	oscill->c_buff->tail = 0;
}

void
a_oscill_get_volt_div(struct a_oscill *oscill, void *data)
{
	return_if_fail(oscill != NULL);

	fprintf(stdout, "Volt division: %lf", oscill->volt_div);
}

void
a_oscill_set_scolor(struct a_oscill *oscill, int scolor)
{
	return_if_fail(oscill != NULL);

	oscill->scolor = scolor;
}

void
a_oscill_get_scolor(struct a_oscill *oscill, void *data)
{
	return_if_fail(oscill != NULL);

	switch(oscill->scolor) {
	case red:
		fprintf(stdout, "Color of screen: red\n");
		break;
	case blue:
		fprintf(stdout, "Color of screen: blue\n");
		break;
	case green:
		fprintf(stdout, "Color of screen: green\n");
		break;
	case yellow:
		fprintf(stdout, "Color of screen: yellow\n");
		break;
	default:
		fprintf(stdout, "Color of screen: white\n");
		break;
	}
}

void
a_oscill_set_ccolor(struct a_oscill *oscill, int ccolor)
{
	return_if_fail(oscill != NULL);

	oscill->ccolor = ccolor;
}

void
a_oscill_get_ccolor(struct a_oscill *oscill, void *data)
{
	return_if_fail(oscill != NULL);

	switch(oscill->ccolor) {
	case red:
		fprintf(stdout, "Color of curve: red\n");
		break;
	case blue:
		fprintf(stdout, "Color of curve: blue\n");
		break;
	case green:
		fprintf(stdout, "Color of curve: green\n");
		break;
	case yellow:
		fprintf(stdout, "Color of curve: yellow\n");
		break;
	default:
		fprintf(stdout, "Color of curve: black\n");
		break;
	}
}

void
a_oscill_set_t_update_screen(struct a_oscill *oscill, short t_screen_update)
{
	return_if_fail(oscill != NULL);

	oscill->t_screen_update = t_screen_update;
}

void
a_oscill_get_t_update_screen(struct a_oscill *oscill, void *data)
{
	return_if_fail(oscill != NULL);

	fprintf(stdout, "Time update screen: %d milliseconds\n", oscill->t_screen_update);
}

/*------------------------function for oscill: end-------------------*/

/*---------------------function for approximation: begin-------------*/
short
a_oscill_convert_signal(double volt, double volt_div)
{
	long n;

	n = (long) (volt / volt_div);

	if (volt / volt_div - n > 0.5)
		n++;

	if (n >= 0) {

		if (n <= SCREEN_HEIGHT / 2)
			return (SCREEN_HEIGHT / 2 - n);
		else
			return -1;
	}

	if (n < 0) {

		if (n >= -SCREEN_HEIGHT / 2)
			return (SCREEN_HEIGHT / 2 + abs(n));
		else
			return SCREEN_HEIGHT + 1;
	}

	return 0;
}

void
a_oscill_linear_approx(struct a_oscill *oscill, double newsample)
{

	if (oscill->c_buff->head == oscill->c_buff->tail) {

		short sample;

		oscill->prev_sample = newsample;

		oscill->last_update_screen = SDL_GetTicks();

		sample = a_oscill_convert_signal(newsample, oscill->volt_div);

		a_oscill_write_c_buff(oscill->c_buff, sample);

		if (oscill->i_time_div >= oscill->time_div) {

			oscill->flag = 0;
			oscill->solve_time = oscill->time_div;

		} else {

			double t = oscill->time_div;

			oscill->flag = 0;

			while (t > oscill->i_time_div) {
				oscill->flag++;
				t = t - oscill->i_time_div;
			}

			oscill->solve_time = t;
		}
		return;
	} else {

		if (oscill->i_time_div >= oscill->time_div) {

			double koeff, t, solve_volt;
			short sample;

			koeff = (newsample - oscill->prev_sample) / oscill->i_time_div;

			t = oscill->solve_time;

			while (t <= oscill->i_time_div) {

				solve_volt = koeff * t + oscill->prev_sample;

				sample = a_oscill_convert_signal(solve_volt, oscill->volt_div);

				a_oscill_write_c_buff(oscill->c_buff, sample);

				t += oscill->time_div;
			}

			oscill->solve_time  = t - oscill->i_time_div;
			oscill->prev_sample = newsample;

		} else {

			if (oscill->flag != 0) {
				oscill->flag--;
				oscill->prev_sample = newsample;
				return;
			} else {
				double koeff, t, solve_volt;

				short sample;

				koeff = (newsample - oscill->prev_sample) / oscill->i_time_div;

				solve_volt = koeff * oscill->solve_time + oscill->prev_sample;

				sample = a_oscill_convert_signal(solve_volt, oscill->volt_div);

				a_oscill_write_c_buff(oscill->c_buff, sample);

				t = oscill->solve_time + oscill->time_div;

				while (t > oscill->i_time_div) {
					oscill->flag++;
					t -= oscill->i_time_div;
				}

				oscill->solve_time = t;

				oscill->prev_sample = newsample;

			}
		}

		if (SDL_GetTicks() - oscill->last_update_screen >= oscill->t_screen_update) {
			a_oscill_update_screen(oscill->screen, oscill->c_buff, oscill->ccolor, oscill->scolor);
			oscill->last_update_screen = SDL_GetTicks();
		}
	}
}

/*---------------------function for approximation: end---------------*/

/*function for interaction with struct a_oscill: begin---------------*/

void
a_oscill_set_time_cmd(struct a_oscill *oscill, const char *data)
{
	return_if_fail(oscill != NULL);

	double time_div;

	while (*data == ' ')
		data++;

	time_div = strtod(data, NULL);

	if (time_div < 0)
		fprintf(stderr, "invalid argument %s\n", data);
	else {
		oscill->time_div = time_div;
		oscill->c_buff->head = 0;
		oscill->c_buff->tail = 0;
	}
}

void
a_oscill_get_time_cmd(struct a_oscill *oscill, const char *data)
{
	return_if_fail(oscill != NULL);

	printf("%lf\n", oscill->time_div);

	if (strlen(data))
		printf("warning: this function not accepts arguments\n");

}

void
a_oscill_set_voltage_cmd(struct a_oscill *oscill, const char *data)
{
	return_if_fail(oscill != NULL);

	double volt_div;

	while (*data == ' ')
		data++;

	volt_div = strtod(data, NULL);

	if (volt_div < 0)
		fprintf(stderr, "invalid argument %s\n", data);
	else {
		oscill->volt_div = volt_div;
		oscill->c_buff->head = 0;
		oscill->c_buff->tail = 0;
	}
}

void
a_oscill_get_voltage_cmd(struct a_oscill *oscill, const char *data)
{
	return_if_fail(oscill != NULL);

	printf("%lf\n", oscill->volt_div);

	if (strlen(data))
		printf("warning: this function not accepts arguments\n");
}

void
a_oscill_set_ccolor_cmd(struct a_oscill *oscill, const char *data)
{
	const char *ptr = data;
	int i;

	while (*ptr == ' ')
		ptr++;

	for (i = 0; i < NUM_COLOR; i++)
		if (!strncmp(color_list[i].namecolor, ptr, strlen(color_list[i].namecolor)))
			oscill->ccolor = color_list[i].numcolor;
}

void a_oscill_get_ccolor_cmd(struct a_oscill *oscill, const char *data)
{
	return_if_fail(oscill != NULL);

	switch(oscill->ccolor) {
		case red:
			printf("red\n");
			break;
		case yellow:
			printf("yellow\n");
			break;
		case blue:
			printf("blue\n");
			break;
		case green:
			printf("green\n");
		default:
			printf("black\n");
			break;
	}

	if (data != NULL)
		if (strlen(data))
			printf("warning: this function not accepts arguments\n");
}

void
a_oscill_set_scolor_cmd(struct a_oscill *oscill, const char *data)
{
	const char *ptr = data;
	int i;

	while (*ptr == ' ')
		ptr++;

	for (i = 0; i < NUM_COLOR; i++)
		if (!strncmp(color_list[i].namecolor, ptr, strlen(color_list[i].namecolor)))
			oscill->scolor = color_list[i].numcolor;
}

void
a_oscill_get_scolor_cmd(struct a_oscill *oscill, const char *data)
{
	return_if_fail(oscill != NULL);

	switch(oscill->scolor) {
		case red:
			printf("red\n");
			break;
		case yellow:
			printf("yellow\n");
			break;
		case blue:
			printf("blue\n");
			break;
		case green:
			printf("green\n");
			break;
		default:
			printf("white\n");
			break;
	}

	if (data != NULL)
		if (strlen(data))
			printf("warning: this function not accepts arguments\n");
}

void
a_oscill_set_i_time_cmd(struct a_oscill *oscill, const char *data)
{
	return_if_fail(oscill != NULL);

	double i_time_div;

	while (*data == ' ')
		data++;

	i_time_div = strtod(data, NULL);

	if (i_time_div < 0)
		fprintf(stderr, "invalid argument %s\n", data);
	else {
		oscill->i_time_div = i_time_div;
		oscill->c_buff->head = 0;
		oscill->c_buff->tail = 0;
	}
}

void
a_oscill_get_i_time_cmd(struct a_oscill *oscill, const char *data)
{
	return_if_fail(oscill != NULL);

	printf("%lf\n", oscill->i_time_div);

	if (strlen(data))
		printf("warning: this function not accepts arguments\n");
}

void
a_oscill_set_t_update_screen_cmd(struct a_oscill *oscill, const char *data)
{
	return_if_fail(oscill != NULL);

	int t_update_screen;

	while (*data == ' ')
		data++;

	t_update_screen = strtol(data, NULL, 10);

	if (t_update_screen < 0)
		fprintf(stderr, "invalid argument %s\n", data);
	else
		oscill->t_screen_update = t_update_screen;
}

void
a_oscill_get_t_update_screen_cmd(struct a_oscill *oscill, const char *data)
{
	return_if_fail(oscill != NULL);

	printf("%d\n", oscill->t_screen_update);

	if (strlen(data))
		printf("warning: this function not accepts arguments\n");
}


/*function for interaction with struct a_oscill: end-----------------*/

int
a_oscill_fread(FILE *fp, char *buff, int sizebuff)
{
	return_val_if_fail(fp != NULL, A_OSCILL_ERR);
	int c;
	int i = 0;

	while ((c = getc(fp)) != EOF && c != '\n' && c != ' ' && i < sizebuff) {
		buff[i] = c;
		i++;
	}

	return i;
}

